/*  Rafa Gwka G1ISI
    Skierniewice 24.03.2008

    ZAD11 RN 
	Wzorzec projektowy proxy, leniwa inicjacja (virtual proxy, tworzy obiekt gdy pierwsze odwoanie).

	   Wzorzec proxy jest wzorcem strukturalnym. Wzorce strukturalne opisuj sposoby czenia klas i obiektw
	w wiksze struktury.  Obiekt proxy jest obiektem poredniczcym miedzy klientem a rzeczywistym obiektem 
	(zwykle potrzebujcym sporych zasobw). Proxy nadzoruje dostp do obiektu i tworzy go dopiero, gdy jest potrzebny.
	Podstawow zaleta pync z jego zastosowania jest tworzenie tylko tych obiektw, ktre faktycznie uywamy. 

	Cechy proxy :
	 -) Taki sam interfejs jak obiekt docelowy
	 -) Proxy przesya otrzymane polecenia do obiektu docelowego.
	 -) Faktyczne tworzenie obiektu w momencie pierwszego uycia. 

    Ilustracja :              Klient - - - - - > Interfejs podmiotu <|---|---------------- PROXY
	                                                                     |                   | delegacja polece
                                                                      Podmiot <--------------|

    adny obrazek na http://en.wikipedia.org/wiki/Image:Proxy_pattern_diagram.png

	Opis przykadu :
	   Do czsto jednym z elementw rozbudowanej gry (lub innej aplikacji graficznej) jest bibliotek modeli.
	 Znajduj si w niej dane o wygldzie (geometrii) wszystkich obiektw wystpujacych w grze.
	 Modele te we wspczesnych grach sa bardzo szczegowe i zajmuj sporo miejsca. 
	 Trzeba zauway ze np w grach typu RTS gracz ma moliwo budowania wielu rodzaji budynkw i jednostek.
	 Jednak wikszo graczy ma wasn strategie gry i uywa tylko czci z nich. Jest to, zatem idealny
	 kontekst do uycia wzorca virtual proxy.

    Wady : 
	-) Due obcienie procesora przy 1 wykorzystaniu obiektu. (W naszym przykadzie chwilowy spadek FPS-w)

	Uwagi :
	-) W programie wielowatkowym naleaoby wprowadzic sekcje krytyczne w chwili tworzenia
	   i  niszczenia realnego obiektu. ( Z uwagi na koniecznosc sprawdzania czy obiekt realny
	   istnieje przy kazdej operacji, byoby to potencjalnie waskie gardlo. )
	   Przykad jest 1 wtkowy, wiec jedynie oznaczyem komentarzem miejsca potencjalnych 
	   sekcji krytycznych.


*/

#include <iostream>
#include <string>

using namespace std;

// Abstrakcyjny interfejs obiektu przechywujacego dane
class Model3D_Interface 
{ 
  public:
	Model3D_Interface(const string& filename){}
	virtual void Render() = 0;
	virtual ~Model3D_Interface(){}
  private :
    Model3D_Interface() {}
};

/*
 Klasa repzentujca model wraz z geometria. Przykad ukazuje sam mechanizm wzorca proxy zatem
 caa nieistotna z tego punktu widzenia zawarto zostaa usunita. 
*/
class Model3D : public Model3D_Interface 
{
  public:
	Model3D(const string& filename) : Model3D_Interface(filename)
	{ 
		cout<< "Wczytanie geometrii modelu " << filename << endl; 
	}
	virtual void Render() {}
	virtual ~Model3D() { /*Kadowanie geometrii modelu*/ }
};

/* 
  Klasa proxy kontrolujaca dostep do waciwej klasy modelu. Sama w sobie zawiera niewiele danych i zajmuje 
  niewiele pamici. Nadzoruje jedynie dostp do obiektu i w momecie pierwszego uycia tworzy realny obiekt. 
  Jest to operacja transparentna dla uytkownika, jednak nalezy liczy si z moliwoci niewielkiego oponienia
  zwiazanego z dostpem do dysku.
*/

class Model3D_Proxy : public Model3D_Interface {
public:
	Model3D_Proxy(const string& filename) : Model3D_Interface(filename)
	{
		filename_=filename;
		realmodel_=NULL; 
        cout<< "Tworzenie porednika modelu " << filename <<endl;
	};

	virtual void Render()
	{
	  // Sterowanie jest przekazywane do rownowanej funkcji obiektu waciwego za pomoca
      // funkcji model ktora w jednolity i srpawny sposob kontroluje istnienie obiektu.
	  model()->Render();
      // Powyszy schemat dziaania wystpuje w kadej metodzie korzystajacej z funkcjonalnoci waciwego obiektu.
	};

	virtual ~Model3D_Proxy() 
	{
	  // Jeeli waciwy obiekt zosta utworzony naley posprzta po sobie. 	  
	  // Potencjalna sekcja krytyczna
	  if (realmodel_!=NULL) delete realmodel_; 
	  realmodel_=NULL; // Zabespieczenie przed wielokrotnym wolaniem destruktora.
	};

private:
	Model3D* realmodel_; // Referencja na obiekt waciwy.
	
	// Dane niezbedne do ponego utworzenia obiektu.
	string filename_; // Nazwa pliku z modelem.

	// Prywatna metoda w elegancji sposob rozwiazujaca koniecznosc kazdorazowego
	// sprawdzania czy obiekt istnieje. Jezeli nie to tworzy go. 
	Model3D* model()
	{
  	   // Potencjalna sekcja krytyczna
	   if (realmodel_==NULL) realmodel_ = new Model3D(filename_); 

	   return realmodel_;
	}

};

int main()
{
   cout << "Inicjalizacja danych i modeli" << endl;

   // Stworzenie namiastki biblioteki jednostek z typowej gry RTS.
   Model3D_Proxy tank("tank.max");
   Model3D_Proxy helicopter("helicopter.max");
   Model3D_Proxy submarine("submarine.max");
   Model3D_Proxy f16("f16.max");

   cout << "Rozpoczecie pracy aplikacji" << endl;

   // Akurat uytkownik prowadzi dziaania na pusty wic korzysta tylko z 2 rodzaji jednostek.
   tank.Render();
   helicopter.Render();

   cout << "Zakoczenie pracy aplikacji" << endl;
   /*
     Wydruk konsoli :
       Inicjalizacja danych i modeli
       Tworzenie porednika modelu tank.max
       Tworzenie porednika modelu helicopter.max
       Tworzenie porednika modelu submarine.max
       Tworzenie porednika modelu f16.max
       Rozpoczecie pracy aplikacji
       Wczytanie geometrii modelu tank.max
       Wczytanie geometrii modelu helicopter.max
       Zakoczenie pracy aplikacji       
     
	 Jak wida wzorzec speni swoje zadanie i zostay utworzone tylko uywane obiekty. 
   */

   return 0;
}